#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import re
import base64
import json
import enum
import pkgutil
import gzip
import time
import textwrap
import datetime
import pprint
import collections
import urllib.parse
import pyasn1.type.univ
import pyasn1.codec.der.encoder
import pyasn1.codec.der.decoder
import pyasn1.codec.ber.decoder
from x509sak.Exceptions import UnexpectedFileContentException, InvalidUsageException, InvalidInputException

class PEMDataTools():
	@classmethod
	def pem2data(cls, pem_text, marker, ignore_errors = False):
		line_begin = "-----BEGIN %s-----" % (marker)
		line_end = "-----END %s-----" % (marker)
		result = [ ]
		data = None
		for line in pem_text.split("\n"):
			line = line.rstrip("\r")
			if data is not None:
				if line == line_end:
					# Finished!
					result.append(base64.b64decode("".join(data)))
					data = None
				else:
					data.append(line)
			else:
				if line == line_begin:
					data = [ ]

		if (len(result) == 0) and (not ignore_errors):
			if data is None:
				raise UnexpectedFileContentException("No begin marker found: Not a %s?" % (marker))
			else:
				raise UnexpectedFileContentException("No end marker found. %s data incomplete?" % (marker))
		return result

	@classmethod
	def data2pem(cls, data, marker):
		if marker is None:
			raise InvalidUsageException("Cannot encode object in PEM without a given marker.")

		line_begin = "-----BEGIN %s-----" % (marker)
		line_end = "-----END %s-----" % (marker)
		lines = [ line_begin ]
		lines += textwrap.wrap(base64.b64encode(data).decode("ascii"), width = 64)
		lines.append(line_end)
		return "\n".join(lines)

class CmdTools():
	_ENV_ALWAYS_EXPORT = [ "SOFTHSM2_CONF", "COVERAGE_FILE" ]

	@classmethod
	def cmdline(cls, cmd, env = None):
		if env is None:
			env = { }
		else:
			env = dict(env)
		for varname in cls._ENV_ALWAYS_EXPORT:
			if (varname in os.environ) and (varname not in env):
				env[varname] = os.environ[varname]

		def escape(text):
			if (" " in text) or ("\"" in text) or ("'" in text) or (";" in text) or ("&" in text) or ("*" in text):
				return "'%s'" % (text.replace("'", "\'"))
			else:
				return text
		command = " ".join(escape(arg) for arg in cmd)

		if env is None:
			return command
		else:
			env_string = " ".join("%s=%s" % (key, escape(value)) for (key, value) in sorted(env.items()))
			return (env_string + " " + command).lstrip()

class ASN1Tools():
	_REGEX_UTCTime = re.compile(r"(?P<year>\d{2})(?P<month>\d{2})(?P<day>\d{2})(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z")
	_REGEX_GeneralizedTime = re.compile(r"(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z")
	_DecodedASN1 = collections.namedtuple("DecodedASN1", [ "asn1", "generic_asn1", "tail", "flags", "original_der", "encoded_der", "model_index" ])

	@classmethod
	def parse_datetime(cls, datetime_str):
		try:
			result = cls._REGEX_UTCTime.fullmatch(datetime_str)
			if result is not None:
				result = { key: int(value) for (key, value) in result.groupdict().items() }
				if result["year"] >= 50:
					result["year"] += 1900
				else:
					result["year"] += 2000
				return datetime.datetime(result["year"], result["month"], result["day"], result["hour"], result["minute"], result["second"])

			result = cls._REGEX_GeneralizedTime.fullmatch(datetime_str)
			if result is not None:
				result = { key: int(value) for (key, value) in result.groupdict().items() }
				return datetime.datetime(result["year"], result["month"], result["day"], result["hour"], result["minute"], result["second"])
		except ValueError:
			pass
		return None

	@classmethod
	def bitstring2bytes(cls, bitstr):
		if (len(bitstr) % 8) != 0:
			raise Exception("Unable to decode ASN.1 BitString to bytes. Size not a multiple of 8: %d" % (len(bitstr)))
		bytes_data = bytearray()
		for i in range(0, len(bitstr), 8):
			byte_data = bitstr[i : i + 8]
			value = sum(bitval << bitpos for (bitpos, bitval) in enumerate(reversed(byte_data)))
			bytes_data.append(value)
		return bytes(bytes_data)

	@classmethod
	def bytes2bitstring(cls, bytedata):
		bitstring = [ ]
		for bytevalue in bytedata:
			for bit in reversed(range(8)):
				bitstring.append((bytevalue >> bit) & 1)
		return pyasn1.type.univ.BitString(bitstring)

	@classmethod
	def bitstring_is_empty(cls, bitstr):
		return (len(bitstr) == 0) or (set(bitstr) == set([ 0 ]))

	@classmethod
	def bitstring_has_trailing_zeros(cls, bitstr):
		return (len(bitstr) > 0) and (bitstr[len(bitstr) - 1] == 0)

	@classmethod
	def bitstring_highbit(cls, bitstr):
		if len(bitstr) == 0:
			return None
		for bit in range(len(bitstr) - 1, -1, -1):
			if bitstr[bit] == 1:
				return bit
		return None

	@classmethod
	def redecode(cls, asn1, asn1_spec):
		encoded_asn1 = pyasn1.codec.der.encoder.encode(asn1)
		return pyasn1.codec.der.decoder.decode(encoded_asn1, asn1Spec = asn1_spec)

	@classmethod
	def safe_decode(cls, der_data, asn1_spec = None):
		def decode_reencode(der_data, asn1_spec = None):
			try:
				(asn1, tail) = pyasn1.codec.ber.decoder.decode(der_data, asn1Spec = asn1_spec)
			except pyasn1.error.PyAsn1Error:
				return None

			try:
				reencoded = pyasn1.codec.der.encoder.encode(asn1)
			except pyasn1.error.PyAsn1Error:
				# Decodable, but not re-encodable!
				reencoded = None
			return (asn1, tail, reencoded)

		result = {
			"asn1": None,
			"generic_asn1": None,
			"tail": None,
			"flags": set(),
			"original_der": der_data,
			"encoded_der": None,
			"model_index": None,
		}

		if isinstance(asn1_spec, tuple):
			asn1_specs = asn1_spec
		else:
			asn1_specs = [ asn1_spec ]

		for (model_index, asn1_spec) in enumerate(asn1_specs):
			decoding_result = decode_reencode(der_data, asn1_spec)
			if decoding_result is not None:
				if model_index != 0:
					result["flags"].add("fallback")
				(result["asn1"], result["tail"], result["encoded_der"]) = decoding_result
				result["model_index"] = model_index
				break
		else:
			# No type matched. Try to decode generically, entirely without asn1Spec
			decoding_result = decode_reencode(der_data)
			if decoding_result is not None:
				(result["generic_asn1"], result["tail"], result["encoded_der"]) = decoding_result
				result["flags"].add("unexpected_type")
			else:
				result["flags"].add("undecodable")

		if (result["tail"] is not None) and (len(result["tail"]) > 0):
			result["flags"].add("trailing_data")
		if result["encoded_der"] is None:
			if not "undecodable" in result["flags"]:
				result["flags"].add("non_encodable")
		else:
			if (result["encoded_der"] + result["tail"]) == der_data:
				# If it's the same after re-encoding memory, we don't need to waste memory
				result["encoded_der"] = None
			else:
				result["flags"].add("non_der")

		return cls._DecodedASN1(**result)

class ECCTools():
	@classmethod
	def decode_enc_pubkey(cls, enc_pubkey):
		if enc_pubkey[0] != 0x04:
			raise Exception("Unable to decode compressed (0x%x) EC key." % (enc_pubkey[0]))
		if (len(enc_pubkey) % 2) != 1:
			raise Exception("Unable to determine correct splitting of %d-bytes inner EC key." % (len(enc_pubkey)))
		bytelen = len(enc_pubkey) // 2
		x = int.from_bytes(enc_pubkey[1 : 1 + bytelen], byteorder = "big")
		y = int.from_bytes(enc_pubkey[1 + bytelen : ], byteorder = "big")
		return (x, y)

class PathTools():
	@classmethod
	def find(cls, search_path, filename):
		directories = search_path.split(":")
		for directory in directories:
			if not directory.endswith("/"):
				directory += "/"
			full_filename = directory + filename
			if os.path.isfile(directory + filename):
				return full_filename
		return None

class DateTimeTools():
	@classmethod
	def strfmt_iso(cls, timestamp):
		"""timestamp.strftime("%Y-%m-%dT%H:%M:%S") produces for years less than
		4 digits no leading zeros. This is something that strptime then cannot
		parse anymore. We therefore implement it ourselves."""
		return "%04d-%02d-%02dT%02d:%02d:%02dZ" % (timestamp.year, timestamp.month, timestamp.day, timestamp.hour, timestamp.minute, timestamp.second)

class JSONTools():
	@classmethod
	def translate(cls, obj):
		if isinstance(obj, dict):
			translated = { key: cls.translate(value) for (key, value) in obj.items() }
		elif isinstance(obj, (list, tuple)):
			translated = [ cls.translate(value) for value in obj ]
		elif isinstance(obj, set):
			translated = sorted(cls.translate(value) for value in obj)
		elif isinstance(obj, enum.IntEnum):
			translated = {
				"name":		obj.name,
				"value":	obj.value,
			}
		elif isinstance(obj, datetime.datetime):
			return obj.strftime("%Y-%m-%d %H:%M:%S")
		else:
			dict_converter = getattr(obj, "to_dict", None)
			if dict_converter is not None:
				translated = cls.translate(dict_converter())
			else:
				translated = obj
		return translated

	class Encoder(json.JSONEncoder):
		def iterencode(self, obj, _one_shot = False):
			# We need to have a pre-translate step because the stupid Python
			# standard encoder supplies no way of intercepting known data types
			# such as int. IntEnum is an instance of int, unfortunately,
			# therefore it's always encoded as a number.
			obj = JSONTools.translate(obj)
			return super().iterencode(obj, _one_shot)

	@classmethod
	def serialize(cls, data):
		return json.dumps(data, indent = 4, sort_keys = True, cls = cls.Encoder)

	@classmethod
	def write_to_fp(cls, data, fp, pretty = False):
		if pretty:
			json.dump(data, fp = fp, indent = 4, sort_keys = True, cls = cls.Encoder)
		else:
			json.dump(data, fp = fp, separators = (',', ':'), cls = cls.Encoder)
		print(file = fp)

	@classmethod
	def write_to_file(cls, data, filename):
		with open(filename, "w") as fp:
			cls.write_to_fp(data, fp)

	@classmethod
	def load_internal(cls, pkgname, filename):
		data = pkgutil.get_data(pkgname, filename)
		if filename.endswith(".gz"):
			data = gzip.decompress(data)
		# Older Python versions (3.5, running on Travis-CI) require this to be
		# str, not bytes.
		data = data.decode("ascii")
		return json.loads(data)

class TextTools():
	@classmethod
	def sp(cls, count, singular, plural = None):
		if plural is None:
			plural = singular + "s"
		if count == 1:
			return "one %s" % (singular)
		else:
			return "%d %s" % (count, plural)

	@classmethod
	def abbreviate(cls, text, to_length):
		if len(text) <= to_length:
			return text
		else:
			textlen = to_length - 3
			tail = round(textlen / 4)
			head = textlen - tail
			if tail + head >= len(text):
				# Only abbreviate tail
				return text[ : (to_length - 3)] + "..."
			else:
				# Abbreviate head and tail
				return text[ : head] + "..." + text[-tail : ]

class ValidationTools():
	_DOMAIN_NAME_RE = re.compile(r"([-a-zA-Z0-9]+\.)*[-a-zA-Z0-9]+")
	_EMAIL_ADDRESS_RE = re.compile(r"(?P<mailbox>[-a-zA-Z0-9.!#$%&'*+/=?^_`{|}~]+)@(?P<domainname>.*)")
	_URI_RE = re.compile(r"(?P<scheme>[a-z]+):(?P<authority>/*[-a-zA-Z0-9+%_.:,=;@\[\]]+)?(?P<path>/[-a-zA-Z0-9+%_.:,=;@/]+)?(?P<query>\?[-a-zA-Z0-9+%_.:,=;@/?#]*)?")
	_LABEL_RE = re.compile(r"(?P<left>[-a-zA-Z0-9]*)(?P<wildcard>\*)?(?P<right>[-a-zA-Z0-9*]*)")

	class DomainnameTemplateValidationResult(enum.IntEnum):
		Valid = 0
		InvalidCharacter = 1
		MoreThanOneWildcard = 2
		FullWildcardNotLeftmost = 3
		WildcardInInternationalDomain = 4

	@classmethod
	def validate_email_address(cls, email_address):
		"""Validate an email address according to a subset of RFC5322; i.e., we
		don't even accept all RFC5322 mail addresses and even less all RFC822
		email addresses here. This means that (technically) valid email
		addresses will fail this check."""
		match = cls._EMAIL_ADDRESS_RE.fullmatch(email_address)
		if match is not None:
			match = match.groupdict()
			return cls.validate_domainname(match["domainname"])
		else:
			return False

	@classmethod
	def validate_domainname(cls, domainname):
		return cls._DOMAIN_NAME_RE.fullmatch(domainname) is not None

	@classmethod
	def validate_domainname_template(cls, domainname_template):
		labels = domainname_template.lower().split(".")
		seen_wildcard = False
		for (lid, label) in enumerate(labels):
			result = cls._LABEL_RE.fullmatch(label)
			if result is None:
				return (cls.DomainnameTemplateValidationResult.InvalidCharacter, label)
			result = result.groupdict()

			if result["wildcard"] is not None:
				if label.startswith("xn--"):
					return (cls.DomainnameTemplateValidationResult.WildcardInInternationalDomain, label)

				if seen_wildcard:
					return (cls.DomainnameTemplateValidationResult.MoreThanOneWildcard, label)
				seen_wildcard = True

				if (result["left"] == "") and (result["right"] == ""):
					if lid != 0:
						return (cls.DomainnameTemplateValidationResult.FullWildcardNotLeftmost, label)
		return (cls.DomainnameTemplateValidationResult.Valid, domainname_template)

	@classmethod
	def _escape_regex(cls, text):
		for escape_char in "\\*.+?()[]":
			text = text.replace(escape_char, "\\" + escape_char)
		return text

	@classmethod
	def validate_domainname_template_match(cls, template, domainname):
		template_re = cls._escape_regex(template.lower()).replace("\\*", r"[-a-z0-9]*")
		match = re.fullmatch(template_re, domainname.lower())
		return match is not None

	@classmethod
	def validate_uri(cls, uri):
		if cls._URI_RE.fullmatch(uri) is None:
			return False
		split_uri = urllib.parse.urlsplit(uri)
		if split_uri.scheme == "":
			return False
		if split_uri.scheme in [ "http", "https" ]:
			return (split_uri.netloc != "") and ((split_uri.path == "") or (split_uri.path.startswith("/")))
		else:
			return True

class PaddingTools():
	@classmethod
	def unpad_pkcs1(cls, data):
		if data[0] != 1:
			raise InvalidInputException("PKCS#1 padding must start with 0x01")

		i = 0
		last_char = None
		for i in range(1, len(data)):
			if data[i] == 0xff:
				continue
			elif data[i] == 0x0:
				# Finished
				last_char = i
				break
			else:
				raise InvalidInputException("PKCS#1 padding must be either 0xff or 0x00 at offset %d, was 0x%02x" % (i, data[i]))

		if last_char is None:
			raise InvalidInputException("PKCS#1 padding does not seem to contain data.")
		return data[i + 1:]

class FileLockTools():
	class _FileLockObject():
		def __init__(self, lock_dir):
			self._lock_dir = lock_dir

		def __enter__(self):
			return self

		def __exit__(self, *args):
			try:
				os.rmdir(self._lock_dir)
			except FileNotFoundError:
				pass

	@classmethod
	def lock(cls, filename, retry = 0.25):
		filename = os.path.realpath(filename)
		lock_name = os.path.dirname(filename) + "/." + os.path.basename(filename) + ".lock"
		while True:
			try:
				os.mkdir(lock_name)
				# Acquired lock
				break
			except FileExistsError:
				# Other process holds lock
				time.sleep(retry)
		return FileLockTools._FileLockObject(lock_name)

class DebugTools():
	@classmethod
	def pprint(cls, data):
		pprint.PrettyPrinter().pprint(data)

class DictTools():
	@classmethod
	def merge(cls, *dicts):
		result = { }
		for single_dict in dicts:
			result.update(dict(single_dict))
		return result

	@classmethod
	def expand_multikeys(cls, multikey_dict):
		result = { }
		for (keys, value) in multikey_dict.items():
			if isinstance(keys, tuple):
				for key in keys:
					result[key] = value
			else:
				result[keys] = value
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, JudgementCode, Verdict, Commonness

class SecurityJudgementTests(BaseTest):
	def test_simple(self):
		judgement = SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar!", bits = 0)
		self.assertEqual(judgement.codeenum, JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent)
		self.assertEqual(judgement.text, "Foo Bar!")
		self.assertEqual(judgement.bits, 0)
		self.assertEqual(judgement.compatibility, None)

		# The following are implied
		self.assertEqual(judgement.verdict, Verdict.NO_SECURITY)
		self.assertEqual(judgement.commonness, Commonness.HIGHLY_UNUSUAL)

		result = judgement.to_dict()
		self.assertEqual(result["code"], judgement.code.name)
		self.assertEqual(result["topic"], judgement.code.topic)
		self.assertEqual(result["short_text"], judgement.code.description)
		self.assertEqual(result["text"], judgement.text)
		self.assertEqual(result["bits"], judgement.bits)

	def test_judgements(self):
		judgements = SecurityJudgements()
		self.assertTrue(judgements.uniform_topic)
		self.assertEqual(len(judgements), 0)

		judgements += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 1", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)
		self.assertTrue(judgements.uniform_topic)
		self.assertEqual(len(judgements), 1)

		judgements += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 2", verdict = Verdict.BROKEN, commonness = Commonness.COMMON)
		self.assertTrue(judgements.uniform_topic)
		self.assertEqual(len(judgements), 2)

		self.assertEqual(judgements.verdict, Verdict.BROKEN)
		self.assertEqual(judgements.commonness, Commonness.UNUSUAL)
		self.assertEqual(judgements.bits, None)

	def test_serialize_judgement(self):
		judgement = SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar!", bits = 0)
		serialized = judgement.to_dict()
		self.assertEqual(serialized["code"], judgement.code.name)
		self.assertEqual(serialized["verdict"]["name"], judgement.verdict.name)
		self.assertEqual(serialized["commonness"]["name"], judgement.commonness.name)
		self.assertNotIn("compatibility", serialized)

		judgement = SecurityJudgement.from_dict(serialized)
		self.assertEqual(judgement.codeenum, JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent)
		self.assertEqual(judgement.text, "Foo Bar!")
		self.assertEqual(judgement.bits, 0)
		self.assertEqual(judgement.verdict, Verdict.NO_SECURITY)
		self.assertEqual(judgement.commonness, Commonness.HIGHLY_UNUSUAL)
		self.assertEqual(judgement.compatibility, None)

	def test_serialize_judgements(self):
		judgements = SecurityJudgements()
		judgements += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 1", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)
		judgements += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 2", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)

		serialized = judgements.to_dict()
		self.assertEqual(serialized["components"][0]["text"], "Foo Bar! 1")
		self.assertEqual(serialized["components"][1]["text"], "Foo Bar! 2")

		judgements = SecurityJudgements.from_dict(serialized)
		self.assertEqual(judgements[0].text, "Foo Bar! 1")
		self.assertEqual(judgements[1].text, "Foo Bar! 2")

	def test_append_judgements(self):
		judgements = SecurityJudgements()
		self.assertEqual(len(judgements), 0)

		judgements += None
		sub = SecurityJudgements()
		sub += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 1", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)
		sub += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 2", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)
		judgements += sub
		self.assertEqual(len(judgements), 2)

		judgements += SecurityJudgement(code = JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, text = "Foo Bar! 3", verdict = Verdict.HIGH, commonness = Commonness.UNUSUAL)
		self.assertEqual(len(judgements), 3)

		self.assertEqual(judgements[0].text, "Foo Bar! 1")
		self.assertEqual(judgements[1].text, "Foo Bar! 2")
		self.assertEqual(judgements[2].text, "Foo Bar! 3")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
from x509sak.SubprocessExecutor import SubprocessExecutor

class BaseAction():
	def __init__(self, cmdname, args):
		self._cmdname = cmdname
		self._args = args

		logging.root.setLevel({
			0:	logging.INFO,
		}.get(self._args.verbose, logging.DEBUG))
		formatter = logging.Formatter("{asctime} [{levelname:.1s}]: {name} {message}", style = "{")

		handler = logging.StreamHandler()
		handler.setLevel(logging.DEBUG)
		handler.setFormatter(formatter)
		logging.root.addHandler(handler)

		if self._args.verbose >= 1:
			SubprocessExecutor.set_failed_verbose()
		if self._args.verbose >= 2:
			SubprocessExecutor.set_all_verbose()
		if self._args.verbose >= 3:
			SubprocessExecutor.pause_after_failed_execution()

		self._log = logging.getLogger("x509sak." + self.__class__.__name__)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest
from x509sak.OpenSSLTools import OpenSSLTools
from x509sak.KeySpecification import KeySpecification
from x509sak.AlgorithmDB import Cryptosystems
from x509sak.PrivateKeyStorage import PrivateKeyStorage, PrivateKeyStorageForm
from x509sak.SubprocessExecutor import SubprocessExecutor

class OpenSSLToolsTests(BaseTest):
	@staticmethod
	def _read_file(filename):
		with open(filename, "rb") as f:
			return f.read()

	def test_gen_privkey_rsa(self):
		input_output_data = [
			(
				{ "bitlen": 512 },
				b"512 bit",
			),
			(
				{ "bitlen": 600 },
				b"600 bit",
			),
		]

		for (input_data, output_data) in input_output_data:
			with tempfile.NamedTemporaryFile(prefix = "privkey_", suffix = ".pem") as f:
				OpenSSLTools.create_private_key(PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = f.name), keyspec = KeySpecification(cryptosystem = Cryptosystems.RSA, parameters = input_data))
				content = self._read_file(f.name)
				self.assertIn(b"BEGIN RSA PRIVATE KEY", content)
				self.assertIn(b"END RSA PRIVATE KEY", content)
				output = SubprocessExecutor([ "openssl", "rsa", "-text" ], stdin = content).run().stdout
				self.assertIn(output_data, output)

	def test_gen_privkey_ecc(self):
		input_output_data = [
			(
				{ "curvename": "secp384r1" },
				(b"384 bit", b"secp384r1"),
			),
			(
				{ "curvename": "secp256r1" },
				(b"256 bit", b"prime256v1"),
			),
		]

		for (input_data, output_data) in input_output_data:
			with tempfile.NamedTemporaryFile(prefix = "privkey_", suffix = ".pem") as f:
				OpenSSLTools.create_private_key(PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = f.name), keyspec = KeySpecification(cryptosystem = Cryptosystems.ECC_ECDSA, parameters = input_data))
				content = self._read_file(f.name)
				self.assertNotIn(b"BEGIN EC PARAMETERS", content)
				self.assertNotIn(b"END EC PARAMETERS", content)
				self.assertIn(b"BEGIN EC PRIVATE KEY", content)
				self.assertIn(b"END EC PRIVATE KEY", content)
				output = SubprocessExecutor([ "openssl", "ec", "-text" ], stdin = content).run().stdout
				self.assertIn(output_data[0], output)
				self.assertIn(output_data[1], output)

	def test_gen_csr(self):
		with tempfile.NamedTemporaryFile(prefix = "privkey_", suffix = ".pem") as privkey_file, tempfile.NamedTemporaryFile(prefix = "csr_", suffix = ".pem") as csr_file:
			OpenSSLTools.create_private_key(PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = privkey_file.name), keyspec = KeySpecification.from_cmdline_str("ecc:secp256r1"))
			private_key_storage = PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = privkey_file.name)

			OpenSSLTools.create_csr(private_key_storage = private_key_storage, csr_filename = csr_file.name, subject_dn = "/CN=Foobar")
			output = SubprocessExecutor([ "openssl", "req", "-text" ], stdin = self._read_file(csr_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE REQUEST", output)
			self.assertIn(b"END CERTIFICATE REQUEST", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertNotIn(b"Requested Extensions:", output)

			OpenSSLTools.create_csr(private_key_storage = private_key_storage, csr_filename = csr_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh" ])
			output = SubprocessExecutor([ "openssl", "req", "-text" ], stdin = self._read_file(csr_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE REQUEST", output)
			self.assertIn(b"END CERTIFICATE REQUEST", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"Requested Extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)

			OpenSSLTools.create_csr(private_key_storage = private_key_storage, csr_filename = csr_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ])
			output = SubprocessExecutor([ "openssl", "req", "-text" ], stdin = self._read_file(csr_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE REQUEST", output)
			self.assertIn(b"END CERTIFICATE REQUEST", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"Requested Extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)

			OpenSSLTools.create_csr(private_key_storage = private_key_storage, csr_filename = csr_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ], subject_alternative_ip_addresses = [ "11.22.33.44", "99.88.77.66", "abcd::9876" ])
			output = SubprocessExecutor([ "openssl", "req", "-text" ], stdin = self._read_file(csr_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE REQUEST", output)
			self.assertIn(b"END CERTIFICATE REQUEST", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"Requested Extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)
			self.assertIn(b"IP Address:11.22.33.44", output)
			self.assertIn(b"IP Address:99.88.77.66", output)
			self.assertIn(b"IP Address:ABCD:0:0:0:0:0:0:9876", output)

			OpenSSLTools.create_csr(private_key_storage = private_key_storage, csr_filename = csr_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ], subject_alternative_ip_addresses = [ "11.22.33.44", "99.88.77.66", "abcd::9876" ], x509_extensions = { "2.3.4.5.6.7": "ASN1:UTF8String:Never gonna give you up" })
			output = SubprocessExecutor([ "openssl", "req", "-text" ], stdin = self._read_file(csr_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE REQUEST", output)
			self.assertIn(b"END CERTIFICATE REQUEST", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"Requested Extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)
			self.assertIn(b"IP Address:11.22.33.44", output)
			self.assertIn(b"IP Address:99.88.77.66", output)
			self.assertIn(b"IP Address:ABCD:0:0:0:0:0:0:9876", output)
			self.assertIn(b"Never gonna give you up", output)

	def test_gen_selfsigned_cert(self):
		with tempfile.NamedTemporaryFile(prefix = "privkey_", suffix = ".pem") as privkey_file, tempfile.NamedTemporaryFile(prefix = "crt_", suffix = ".pem") as certificate_file:
			OpenSSLTools.create_private_key(PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = privkey_file.name), keyspec = KeySpecification(cryptosystem = Cryptosystems.ECC_ECDSA, parameters = { "curvename": "secp256r1" }))
			private_key_storage = PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = privkey_file.name)

			OpenSSLTools.create_selfsigned_certificate(private_key_storage = private_key_storage, certificate_filename = certificate_file.name, subject_dn = "/CN=Foobar", validity_days = 365)
			output = SubprocessExecutor([ "openssl", "x509", "-text" ], stdin = self._read_file(certificate_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE--", output)
			self.assertIn(b"END CERTIFICATE--", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertNotIn(b"X509v3 extensions:", output)

			OpenSSLTools.create_selfsigned_certificate(private_key_storage = private_key_storage, certificate_filename = certificate_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh" ], validity_days = 365)
			output = SubprocessExecutor([ "openssl", "x509", "-text" ], stdin = self._read_file(certificate_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE--", output)
			self.assertIn(b"END CERTIFICATE--", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"X509v3 extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)

			OpenSSLTools.create_selfsigned_certificate(private_key_storage = private_key_storage, certificate_filename = certificate_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ], validity_days = 365)
			output = SubprocessExecutor([ "openssl", "x509", "-text" ], stdin = self._read_file(certificate_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE--", output)
			self.assertIn(b"END CERTIFICATE--", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"X509v3 extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)

			OpenSSLTools.create_selfsigned_certificate(private_key_storage = private_key_storage, certificate_filename = certificate_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ], subject_alternative_ip_addresses = [ "11.22.33.44", "99.88.77.66", "abcd::9876" ], validity_days = 365)
			output = SubprocessExecutor([ "openssl", "x509", "-text" ], stdin = self._read_file(certificate_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE--", output)
			self.assertIn(b"END CERTIFICATE--", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"X509v3 extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)
			self.assertIn(b"IP Address:11.22.33.44", output)
			self.assertIn(b"IP Address:99.88.77.66", output)
			self.assertIn(b"IP Address:ABCD:0:0:0:0:0:0:9876", output)

			OpenSSLTools.create_selfsigned_certificate(private_key_storage = private_key_storage, certificate_filename = certificate_file.name, subject_dn = "/CN=Foobar", subject_alternative_dns_names = [ "muhkuh", "kruckelmuckel" ], subject_alternative_ip_addresses = [ "11.22.33.44", "99.88.77.66", "abcd::9876" ], x509_extensions = { "2.3.4.5.6.7": "ASN1:UTF8String:Never gonna give you up" }, validity_days = 365)
			output = SubprocessExecutor([ "openssl", "x509", "-text" ], stdin = self._read_file(certificate_file.name)).run().stdout
			self.assertIn(b"BEGIN CERTIFICATE--", output)
			self.assertIn(b"END CERTIFICATE--", output)
			self.assertTrue((b"Subject: CN = Foobar" in output) or (b"Subject: CN=Foobar" in output))
			self.assertIn(b"X509v3 extensions:", output)
			self.assertIn(b"X509v3 Subject Alternative Name:", output)
			self.assertIn(b"DNS:muhkuh", output)
			self.assertIn(b"DNS:kruckelmuckel", output)
			self.assertIn(b"IP Address:11.22.33.44", output)
			self.assertIn(b"IP Address:99.88.77.66", output)
			self.assertIn(b"IP Address:ABCD:0:0:0:0:0:0:9876", output)
			self.assertIn(b"Never gonna give you up", output)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2021 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
from x509sak.BaseAction import BaseAction
from x509sak import X509Certificate
from x509sak.Tools import JSONTools
from x509sak.FileWriter import FileWriter
from x509sak.OpenSSLTools import OpenSSLTools
from x509sak.CertificateAnalyzer import CertificateAnalyzer
from x509sak.AnalysisPrinter import AnalysisPrinterText

class ActionExamineCert(BaseAction):
	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)
		if self._args.parent_certificate is None:
			parent_ca_cert = None
		else:
			if self._args.in_format == "dercrt":
				# CA cert in DER form if host certificate is also in DER form
				crt = X509Certificate.read_derfile(self._args.parent_certificate)
			else:
				# CA cert in PEM form for all other cases
				crt = X509Certificate.read_pemfile(self._args.parent_certificate)[0]
			parent_ca_cert = CertificateAnalyzer.CertSource(source = self._args.parent_certificate, source_type = "pemcrt" if (self._args.in_format != "dercrt") else "dercrt", crts = [ crt ])

		if self._args.in_format in [ "pemcrt", "dercrt", "host" ]:
			crt_sources = self._load_certificates()

			analysis_params = {
				"include_raw_data":	self._args.include_raw_data,
				"fast_rsa":			self._args.fast_rsa,
				"purposes":			self._args.purpose,
				"entity_name":		self._args.server_name,
			}
			if (len(analysis_params["purposes"]) == 0) and (self._args.in_format == "host") and (not self._args.no_automatic_host_check):
				analysis_params["purposes"].append("tls-server")
			if (analysis_params["entity_name"] is None) and (self._args.in_format == "host") and (not self._args.no_automatic_host_check):
				hostname = crt_sources[0].source.split(":")[0]
				analysis_params["entity_name"] = hostname
			crt_analyzer = CertificateAnalyzer(**analysis_params)

			analysis = crt_analyzer.analyze(crt_sources, parent_ca_cert)
		elif self._args.in_format == "json":
			analysis = self._read_json()
		else:
			raise NotImplementedError(self._args.in_format)

		output = self._args.output or "-"
		with FileWriter(output, "w") as f:
			self._show_analysis(f, analysis)

	def _load_certificates(self):
		sources = [ ]
		for crt_filename in self._args.infiles:
			if self._args.in_format == "pemcrt":
				self._log.debug("Reading PEM certificate from %s", crt_filename)
				crts = X509Certificate.read_pemfile(crt_filename)
			elif self._args.in_format == "dercrt":
				self._log.debug("Reading DER certificate from %s", crt_filename)
				crts = [ X509Certificate.read_derfile(crt_filename) ]
			elif self._args.in_format == "host":
				host_port = crt_filename.split(":", maxsplit = 1)
				if len(host_port) == 1:
					host_port.append("443")
				(host, port) = host_port
				port = int(port)
				self._log.debug("Querying TLS server at %s port %d", host, port)
				crts = [ X509Certificate.from_tls_server(host, port) ]
			else:
				raise NotImplementedError(self._args.in_format)
			source = CertificateAnalyzer.CertSource(source = crt_filename, crts = crts, source_type = self._args.in_format)
			sources.append(source)
		return sources

	def _read_json(self):
		merged_analyses = None
		for json_filename in self._args.infiles:
			with open(json_filename) as f:
				analyses = json.load(f)
				if merged_analyses is None:
					merged_analyses = analyses
				else:
					merged_analyses["data"] += analyses["data"]
		return merged_analyses

	def _show_analysis(self, output, analyses):
		if self._args.out_format in [ "ansitext", "text" ]:
			use_ansi = (self._args.out_format == "ansitext")
			AnalysisPrinterText(output, analyses).print(use_ansi = use_ansi)
		elif self._args.out_format == "json":
			JSONTools.write_to_fp(analyses, output, pretty = self._args.pretty_json)
		else:
			raise NotImplementedError(self._args.out_format)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.FlagChecker import FlagChecker

class FlagCheckerTests(BaseTest):
	def test_all_ok(self):
		checker = FlagChecker().may_have("a", "b", "c")
		self.assertEqual(list(checker.check([ ])), [ ])
		self.assertEqual(list(checker.check([ "a", "c" ])), [ ])
		self.assertEqual(list(checker.check([ "a", "b", "c" ])), [ ])

	def test_missing(self):
		checker = FlagChecker().may_have("a", "b", "c").must_have("d")
		results = list(checker.check([ "a", "c" ]))
		self.assertEqual(len(results), 1)
		self.assertEqual(results[0].check_type, "missing")
		self.assertEqual(list(results[0].flags), [ "d" ])

	def test_unusual(self):
		checker = FlagChecker().may_have("a", "b", "c")
		results = list(checker.check([ "x" ]))
		self.assertEqual(len(results), 1)
		self.assertEqual(results[0].check_type, "unusual")
		self.assertEqual(list(results[0].flags), [ "x" ])

	def test_excessive(self):
		checker = FlagChecker().may_have("a", "b", "c").may_not_have("d")
		results = list(checker.check([ "a", "c" ]))
		self.assertEqual(len(results), 0)

		results = list(checker.check([ "a", "d", "c" ]))
		self.assertEqual(results[0].check_type, "excess")
		self.assertEqual(list(results[0].flags), [ "d" ])

	def test_complex(self):
		checker = FlagChecker().complex_check([ "a", "b", "c"], min_count = 1).may_have("d", "e", "f")
		self.assertEqual(list(checker.check([ "a" ])), [ ])
		self.assertEqual(list(checker.check([ "a", "b" ])), [ ])
		self.assertEqual(list(checker.check([ "a", "b", "c" ])), [ ])
		self.assertEqual(list(checker.check([ "a", "b", "c" ])), [ ])
		self.assertEqual(list(checker.check([ "e", "f", "a" ])), [ ])

		results = list(checker.check([ "e", "f" ]))
		self.assertEqual(len(results), 1)
		self.assertEqual(results[0].check_type, "complex_too_few")
		self.assertEqual(list(results[0].flags), [ ])
		self.assertEqual(list(sorted(results[0].reference)), [ "a", "b", "c" ])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest
from x509sak.WorkDir import WorkDir
from x509sak.X509Certificate import X509Certificate
from x509sak.OID import OIDDB

class CmdLineTestsForgeCert(BaseTest):
	def _assertCrtsSimilar(self, original, forgery):
		self.assertEqual(original.subject, forgery.subject)
		self.assertEqual(original.issuer, forgery.issuer)
		self.assertEqual(original.valid_not_before, forgery.valid_not_before)
		self.assertEqual(original.valid_not_after, forgery.valid_not_after)
		self.assertEqual(original.signature_alg_oid, forgery.signature_alg_oid)
		self.assertEqual(original.pubkey.keyspec, forgery.pubkey.keyspec)
		self.assertNotEqual(original.pubkey, forgery.pubkey)

	def test_forge_root(self):
		root_crt = self._load_crt("ok/johannes-bauer-root")
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			root_crt.write_pemfile("root.crt")
			self._run_x509sak([ "forgecert", "root.crt" ])

			forged_root_crt = X509Certificate.read_pemfile("forged_00.crt")[0]
			self._assertCrtsSimilar(root_crt, forged_root_crt)
			orig_ski = root_crt.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"))
			forged_ski = forged_root_crt.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"))
			self.assertEqual(orig_ski, forged_ski)

	def test_forge_root_new_ski(self):
		root_crt = self._load_crt("ok/johannes-bauer-root")
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			root_crt.write_pemfile("root.crt")
			self._run_x509sak([ "forgecert", "--recalculate-keyids", "root.crt" ])

			forged_root_crt = X509Certificate.read_pemfile("forged_00.crt")[0]
			self._assertCrtsSimilar(root_crt, forged_root_crt)
			orig_ski = root_crt.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"))
			forged_ski = forged_root_crt.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"))
			self.assertNotEqual(orig_ski, forged_ski)

	def test_forge_chain(self):
		root_crt = self._load_crt("ok/johannes-bauer-root")
		intermediate_crt = self._load_crt("ok/johannes-bauer-intermediate")
		server_crt = self._load_crt("ok/johannes-bauer.com")
		orig_crts = [ root_crt, intermediate_crt, server_crt ]

		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			with open("chain.crt", "w") as f:
				for orig_crt in orig_crts:
					print(orig_crt.to_pem_data(), file = f)
			self._run_x509sak([ "forgecert", "chain.crt" ])

			for (num, orig_crt) in enumerate(orig_crts):
				forged_crt = X509Certificate.read_pemfile("forged_%02d.crt" % (num))[0]
				self._assertCrtsSimilar(orig_crt, forged_crt)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.CurveDB import CurveDB
from x509sak.Exceptions import InvalidInputException, UnsupportedEncodingException

class ECCMathTests(BaseTest):
	def test_ecc_prime_curve(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		self.assertTrue(curve.G.on_curve)

	def test_ecc_binary_trinomial_basis_poly(self):
		curve = CurveDB().instantiate(name = "sect113r1")
		self.assertEqual(sorted(curve.poly), [ 0, 9, 113 ])

	def test_ecc_binary_pentanomial_basis_poly(self):
		curve = CurveDB().instantiate(name = "sect163r1")
		self.assertEqual(sorted(curve.poly), [ 0, 3, 6, 7, 163 ])

	def test_ecc_curvedb(self):
		db = CurveDB()
		for curve_oid in db:
			curve = db.instantiate(oid = curve_oid)
			self.assertTrue(curve.G.on_curve())

	def test_point_encode(self):
		point = CurveDB().instantiate(name = "sect113r1").G
		self.assertEqual(point.encode(), bytes.fromhex("04 009d73616f35f4ab1407d73562c10f 00a52830277958ee84d1315ed31886"))

	def test_point_decode(self):
		curve = CurveDB().instantiate(name = "sect113r1")
		decoded_point = curve.decode_point(bytes.fromhex("04 009d73616f35f4ab1407d73562c10f 00a52830277958ee84d1315ed31886"))
		self.assertEqual(decoded_point, curve.G)

	def test_point_decode_fail(self):
		curve = CurveDB().instantiate(name = "sect113r1")
		with self.assertRaises(InvalidInputException):
			curve.decode_point(bytes.fromhex("04 0011 2233"))

		with self.assertRaises(UnsupportedEncodingException):
			curve.decode_point(bytes.fromhex("02 0011 2233"))

	def test_point_str(self):
		point = CurveDB().instantiate(name = "sect113r1").G
		point_str = str(point)
		self.assertIn("0x9d7361", point_str)
		self.assertIn("0xa52830", point_str)
		self.assertIn("sect113r1", point_str)

	def test_point_decode_ed25519(self):
		curve = CurveDB().instantiate(name = "ed25519")

		Q = curve.decode_point(bytes(32))
		self.assertEqual(Q.x, 0x2b8324804fc1df0b2b4d00993dfbd7a72f431806ad2fe478c4ee1b274a0ea0b0)
		self.assertEqual(Q.y, 0)
		self.assertTrue(Q.on_curve())
		self.assertEqual(Q.encode(), bytes(32))

		Q = curve.decode_point(bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000080"))
		self.assertEqual(Q.x, 0x547cdb7fb03e20f4d4b2ff66c2042858d0bce7f952d01b873b11e4d8b5f15f3d)
		self.assertEqual(Q.y, 0)
		self.assertTrue(Q.on_curve())
		self.assertEqual(Q.encode(), bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000080"))

		Q = curve.decode_point(bytes.fromhex("d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"))
		self.assertEqual(Q.x, 0x55d0e09a2b9d34292297e08d60d0f620c513d47253187c24b12786bd777645ce)
		self.assertEqual(Q.y, 0x1a5107f7681a02af2523a6daf372e10e3a0764c9d3fe4bd5b70ab18201985ad7)
		self.assertTrue(Q.on_curve())
		self.assertEqual(Q.encode(), bytes.fromhex("d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"))

		Q = curve.decode_point(bytes.fromhex("dfc9425e4f968f7f0c29f0259cf5f9aed6851c2bb4ad8bfb860cfee0ab248292"))
		self.assertEqual(Q.x, 0x3493c89a1d42961795326fb77ddda9b1073eb50954eec3acc573cd718bed3093)
		self.assertEqual(Q.y, 0x128224abe0fe0c86fb8badb42b1c85d6aef9f59c25f0290c7f8f964f5e42c9df)
		self.assertTrue(Q.on_curve())
		self.assertEqual(Q.encode(), bytes.fromhex("dfc9425e4f968f7f0c29f0259cf5f9aed6851c2bb4ad8bfb860cfee0ab248292"))

		Q = curve.decode_point(bytes.fromhex("dcbfc4d2bd9b5b9b3f7cd673cf559fe3793946a6a904355c07a552991bdba7c5"))
		self.assertEqual(Q.x, 0x5c71bfc23d23bb896be916c12e2b02aa5d22c1883ac097fe5ab604aa52020ec9)
		self.assertEqual(Q.y, 0x45a7db1b9952a5075c3504a9a6463979e39f55cf73d67c3f9b5b9bbdd2c4bfdc)
		self.assertTrue(Q.on_curve())

		with self.assertRaises(InvalidInputException):
			curve.decode_point(bytes.fromhex("0305334e381af78f141cb666f6199f57bc3495335a256a95bd2a55bf546663f6"))

	def test_scalar_mul_ed25519(self):
		curve = CurveDB().instantiate(name = "ed25519")
		self.assertEqual(curve.G.scalar_mul(0), curve.point(0, 1))
		self.assertEqual(curve.G.scalar_mul(1), curve.G)
		self.assertEqual(curve.G, curve.point(0x216936d3cd6e53fec0a4e231fdd6dc5c692cc7609525a7b2c9562d608f25d51a, 0x6666666666666666666666666666666666666666666666666666666666666658))
		self.assertEqual(curve.G.scalar_mul(2), curve.point(0x36ab384c9f5a046c3d043b7d1833e7ac080d8e4515d7a45f83c5a14e2843ce0e, 0x2260cdf3092329c21da25ee8c9a21f5697390f51643851560e5f46ae6af8a3c9))
		self.assertEqual(curve.G.scalar_mul(3), curve.point(0x67ae9c4a22928f491ff4ae743edac83a6343981981624886ac62485fd3f8e25c, 0x1267b1d177ee69aba126a18e60269ef79f16ec176724030402c3684878f5b4d4))
		self.assertEqual(curve.G.scalar_mul(123456789), curve.point(0x547df969eeaad777ccc47f172eb04d76d148ac6fe7e6f03c5f764f1e15327545, 0x5bd3c1a4f2053b458e38123b41e36ddeb5d13a6f63365d93e90ddc6880adff17))

	def test_secret_expand_ed25519(self):
		curve = CurveDB().instantiate(name = "ed25519")
		(scalar, Q) = curve.expand_secret(bytes.fromhex("9d61b19deffd5a60ba844af492ec2cc44449c5697b326919703bac031cae7f60"))
		self.assertEqual(scalar, 36144925721603087658594284515452164870581325872720374094707712194495455132720)
		self.assertEqual(Q.encode(), bytes.fromhex("d75a980182b10ab7d54bfed3c964073a0ee172f3daa62325af021a68f707511a"))

	def test_ed448_generator(self):
		curve = CurveDB().instantiate(name = "ed448")
		rfc_8032_G = curve.point(224580040295924300187604334099896036246789641632564134246125461686950415467406032909029192869357953282578032075146446173674602635247710, 298819210078481492676017930443930673437544040154080242095928241372331506189835876003536878655418784733982303233503462500531545062832660)
		self.assertTrue(rfc_8032_G.on_curve())
		self.assertEqual(curve.G, rfc_8032_G)

	def test_secret_expand_ed448(self):
		curve = CurveDB().instantiate(name = "ed448")
		(scalar, Q) = curve.expand_secret(bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b"))
		self.assertEqual(scalar, 521658399617511624509929819094270498323007786671637499019582168374758478770958028340603419308639592898868374490003595203618871291427304)
		self.assertEqual(Q.encode(), bytes.fromhex("5fd7449b59b461fd2ce787ec616ad46a1da1342485a70e1f8a0ea75d80e96778edf124769b46c7061bd6783df1e50f6cd1fa1abeafe8256180"))

	def test_fp_scalar_mul(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		P = curve.G.scalar_mul(0xa14b08db884ecd9acc3e507110be)
		self.assertEqual(P.x, 0x5d39a5c8d8f5c634afea9d0adf23)
		self.assertEqual(P.y, 0x88056785c1ea5bb9f320eefd630e)

	def test_fp_scalar_mul_n(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		P = curve.G.scalar_mul(curve.n)
		self.assertEqual(P, curve.neutral_point)

	def test_fp_scalar_mul_0(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		P = curve.G.scalar_mul(0)
		self.assertEqual(P, curve.neutral_point)

	def test_fp_scalar_mul_1(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		P = curve.G.scalar_mul(1)
		self.assertEqual(P, curve.G)

	def test_fp_point_dbl(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		P = curve.point_addition(curve.G, curve.G)
		self.assertEqual(P.x, 0x57cf52a0f9318000ee0bc032d756)
		self.assertEqual(P.y, 0x60aee03bbcff537a8d17401f006c)

	def test_fp_point_add(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		Q = curve.point(0x123, 0x84bdce9a00a1895369a805a6c44e)
		self.assertTrue(Q.on_curve())
		P = curve.point_addition(curve.G, Q)
		self.assertEqual(P.x, 0xbbabcf20193b825046cb2357bb87)
		self.assertEqual(P.y, 0x5625e546a0459574b5eff88d17b9)

	def test_fp_point_add_neutral(self):
		curve = CurveDB().instantiate(name = "secp112r1")
		Q = curve.point(0x9487239995a5ee76b55f9c2f098, 0x32df450fdbbe9dc44268aeb5ab8b)
		self.assertTrue(Q.on_curve())
		P = curve.point_addition(curve.G, Q)
		self.assertEqual(P, curve.neutral_point)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate.Judgement import Commonness, Compatibility
from x509sak.estimate.Validator import BaseValidationResult, BaseValidator

class DERValidationResult(BaseValidationResult):
	def _validate(self):
		if "undecodable" in self._subject.flags:
			self._report("Enc_DER_EncodingIssues_Malformed_Undecodable", "input cannot be decoded as valid ASN.1.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
		else:
			if ("non_der" in self._subject.flags) or ("non_encodable" in self._subject.flags):
				suffix = "; re-encoding the ASN.1 structure is not possible" if ("non_encodable" in self._subject.flags) else ""
				self._report("Enc_DER_EncodingIssues_Malformed_NonDEREncoding", "input can be decoded, but does not comply with Distinguished Encoding Rules (DER)%s." % (suffix), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
			if "trailing_data" in self._subject.flags:
				self._report("Enc_DER_EncodingIssues_TrailingData", "input can de decoded as ASN.1 structure, but has %d bytes of trailing data." % (len(self._subject.tail)), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
			if "unexpected_type" in self._subject.flags:
				self._report("Enc_DER_EncodingIssues_Malformed_UnexpectedType", "input can be decoded as ASN.1 structure, but not under the expected schema (decoded as type %s)." % (type(self._subject.generic_asn1).__name__), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

class DERValidator(BaseValidator):
	_ValidationResultClass = DERValidationResult



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
from x509sak.tls.DataBuffer import DataBuffer, DataBufferException
from x509sak.Exceptions import ProgrammerErrorException, InvalidInputException

class StructureException(Exception): pass

class SerializationException(StructureException): pass
class DeserializationException(StructureException): pass

class InvalidInputTypeException(SerializationException): pass
class NoPossibleSerializationException(SerializationException): pass

class NoMoreDataToUnpackException(DeserializationException): pass
class InfiniteUnpackingException(DeserializationException): pass
class IncompleteUnpackingException(DeserializationException): pass
class UnexpectedFixedValueException(DeserializationException): pass
class NoPossibleDeserializationException(DeserializationException): pass

class BaseStructureMember():
	def __init__(self, name = None):
		self._name = name

	@property
	def name(self):
		return self._name

	def pack(self, values):
		raise NotImplementedError(self.__class__.__name__)

	def _unpack(self, databuffer):
		raise NotImplementedError(self.__class__.__name__)

	def unpack(self, databuffer):
		try:
			return self._unpack(databuffer)
		except DataBufferException as e:
			raise NoMoreDataToUnpackException(e)

class StructureMemberFactory():
	_REGISTERED = [ ]

	def __init__(self):
		pass

	@classmethod
	def register(cls, element_class):
		regex = re.compile(element_class._REGEX)
		cls._REGISTERED.append((regex, element_class))
		return element_class

	@classmethod
	def instantiate(cls, name, typename, **kwargs):
		for (regex, element_class) in cls._REGISTERED:
			match = regex.fullmatch(typename)
			if match:
				match = match.groupdict()
				return element_class.from_match(name, match, **kwargs)
		raise ProgrammerErrorException("No handler for type '%s' found." % (typename))

class StructureMemberFactoryElement(BaseStructureMember):
	_REGEX = None

	@classmethod
	def from_match(cls, name, match, **kwargs):
		raise NotImplementedError(cls.__name__)

	@property
	def typename(self):
		return None

	@property
	def implicit_value(self):
		return False

	def __repr__(self):
		cname = self.typename if (self.typename is not None) else self.__class__.__name__
		return "%s %s" % (cname, self.name)

@StructureMemberFactory.register
class StructureElementFixed(StructureMemberFactoryElement):
	_REGEX = r"fixed\[(?P<hexdata>[0-9a-fA-F\s]+)\]"

	def __init__(self, name, fixed_data):
		StructureMemberFactoryElement.__init__(self, name)
		self._fixed_data = fixed_data

	@classmethod
	def from_match(cls, name, match, **kwargs):
		fixed_data = bytes.fromhex(match["hexdata"])
		return cls(name = name, fixed_data = fixed_data, **kwargs)

	@property
	def typename(self):
		return "fixed[%s]" % (self._fixed_data.hex())

	@property
	def implicit_value(self):
		return True

	def _unpack(self, databuffer):
		data = databuffer.get(len(self._fixed_data))
		if data != self._fixed_data:
			raise UnexpectedFixedValueException("%s unpacking expected %s but got %s." % (str(self), self._fixed_data.hex(), data.hex()))
		return data

	def pack(self):
		return self._fixed_data

@StructureMemberFactory.register
class StructureElementInteger(StructureMemberFactoryElement):
	_REGEX = r"uint(?P<bit>\d+)"

	def __init__(self, name, length_bytes, enum_class = None, strict_enum = False, fixed_value = None):
		StructureMemberFactoryElement.__init__(self, name)
		self._length_bytes = length_bytes
		self._enum_class = enum_class
		self._strict_enum = strict_enum
		self._fixed_value = fixed_value
		self._minval = 0
		self._maxval = (1 << (8 * self._length_bytes)) - 1

	@classmethod
	def from_match(cls, name, match, **kwargs):
		length_bits = int(match["bit"])
		assert((length_bits % 8) == 0)
		length_bytes = length_bits // 8
		return cls(name = name, length_bytes = length_bytes, **kwargs)

	@property
	def typename(self):
		return "uint%d" % (self._length_bytes * 8)

	@property
	def implicit_value(self):
		return self._fixed_value is not None

	def _unpack(self, databuffer):
		data = databuffer.get(self._length_bytes)
		value = int.from_bytes(data, byteorder = "big")
		if self._enum_class is not None:
			try:
				value = self._enum_class(value)
			except ValueError:
				if self._strict_enum:
					raise
		if self._fixed_value and (value != self._fixed_value):
			raise UnexpectedFixedValueException("Deserialization requires fixed value 0x%x but encountered 0x%x." % (self._fixed_value, value))
		return value

	def pack(self, value = None):
		if (value is None) and (self._fixed_value is not None):
			value = self._fixed_value
		if not isinstance(value, int):
			raise InvalidInputTypeException("%s requires int to be supplied for packing, got %s: %s" % (self.typename, type(value).__name__, str(value)))
		if self._enum_class is not None:
			if self._strict_enum:
				if not isinstance(value, self._enum_class):
					raise InvalidInputException("%s packing input must be of type %s." % (str(self), self._enum_class))
			value = int(value)
		if (value < self._minval) or (value > self._maxval):
			raise InvalidInputException("%s must be between %d and %d (given value was %d)." % (str(self), self._minval, self._maxval, value))
		data = int.to_bytes(value, byteorder = "big", length = self._length_bytes)
		return data

@StructureMemberFactory.register
class StructureElementOpaque(StructureMemberFactoryElement):
	_REGEX = r"opaque(?P<bit>\d+)"

	def __init__(self, name, length_field, inner = None, contains_array = None, string_encoding = None, fixed_value = None, allow_trailing_data = False):
		StructureMemberFactoryElement.__init__(self, name)
		self._length_field = length_field
		self._inner = inner
		self._contains_array = contains_array
		self._string_encoding = string_encoding
		self._fixed_value = fixed_value
		self._allow_trailing_data = allow_trailing_data
		if (self._string_encoding is not None) and (self._inner is not None):
			raise ProgrammerErrorException("Opaque object can either encode strings or have an inner object, but not both.")

	@classmethod
	def from_match(cls, name, match, **kwargs):
		length_field = StructureMemberFactory.instantiate(name = "length", typename = "uint" + match["bit"])
		return cls(name, length_field = length_field, **kwargs)

	@property
	def typename(self):
		return "opaque<%s>" % (self._length_field)

	@property
	def implicit_value(self):
		return self._fixed_value is not None

	def _unpack(self, databuffer):
		length = self._length_field.unpack(databuffer)
		data = databuffer.get(length)
		if (self._string_encoding is not None):
			data = data.decode(self._string_encoding)
		elif self._inner is not None:
			db = DataBuffer(data)
			if not self._contains_array:
				data = self._inner.unpack(db)
			else:
				result_array = [ ]
				while db.remaining > 0:
					pre_unpack = db.remaining
					result_array.append(self._inner.unpack(db))
					if pre_unpack == db.remaining:
						raise InfiniteUnpackingException("Inner unpacking of %s consumed no data, still %d bytes remaining. Breaking infinite loop." % (str(self._inner), db.remaining))
				data = result_array
			if (db.remaining > 0) and (not self._allow_trailing_data):
				raise IncompleteUnpackingException("%s unpacking still has %d bytes of trailing, non-consumed data left." % (self.typename, db.remaining))
		return data

	def pack(self, data = None):
		if (data is None) and (self._fixed_value is not None):
			data = self._fixed_value

		if (self._string_encoding is not None):
			data = data.encode(self._string_encoding)
		elif self._inner is not None:
			if not self._contains_array:
				data = self._inner.pack(data)
			else:
				packed_data = bytearray()
				for element in data:
					packed_data += self._inner.pack(element)
				data = packed_data

		if not isinstance(data, (bytes, bytearray)):
			raise InvalidInputTypeException("%s requires bytes/bytearray to be supplied for packing, got %s: %s" % (self.typename, type(data).__name__, str(data)))
		return self._length_field.pack(len(data)) + data

@StructureMemberFactory.register
class StructureElementArray(StructureMemberFactoryElement):
	_REGEX = r"array\[(?P<length>\d+)(,\s+(?P<padbyte>[0-9a-fA-F]{2}))?\]"

	def __init__(self, name, length, padbyte = None):
		StructureMemberFactoryElement.__init__(self, name)
		self._length = length
		self._padbyte = padbyte

	@classmethod
	def from_match(cls, name, match, **kwargs):
		length = int(match["length"])
		if match["padbyte"] is not None:
			padbyte = int(match["padbyte"], 16)
		else:
			padbyte = None
		return cls(name, length = length, padbyte = padbyte, **kwargs)

	@property
	def typename(self):
		if self._padbyte is None:
			return "array[%d]" % (self._length)
		else:
			return "array[%d, %02x]" % (self._length, self._padbyte)

	def _unpack(self, databuffer):
		return databuffer.get(self._length)

	def pack(self, data):
		if self._padbyte is None:
			# Size must exactly match up
			if len(data) == self._length:
				return data
			else:
				raise InvalidInputException("For packing of array of length %d without padding, %d bytes must be provided. Got %d bytes." % (self._length, self._length, len(data)))
		else:
			# Can pad
			if len(data) <= self._length:
				pad_len = self._length - len(data)
				padding = bytes([ self._padbyte ]) * pad_len
				return data + padding
			else:
				raise InvalidInputException("For packing of array of length %d with padding, at most %d bytes must be provided. Got %d bytes." % (self._length, self._length, len(data)))

class Structure(BaseStructureMember):
	def __init__(self, members, name = None):
		BaseStructureMember.__init__(self, name = name)
		self._members = members
		self._required_keys = set(member.name for member in members if (not member.implicit_value))
		explicit_member_count = sum(1 for member in members if (not member.implicit_value))
		if len(self._required_keys) != explicit_member_count:
			raise ProgrammerErrorException("Structure definition amgiguous, duplicate member names used.")

	@property
	def typename(self):
		return "Structure"

	@property
	def implicit_value(self):
		return False

	@property
	def members(self):
		return iter(self._members)

	def pack(self, values):
		if not isinstance(values, dict):
			raise InvalidInputTypeException("%s requires a dict to be supplied for packing, got %s: %s" % (self.typename, type(values).__name__, str(values)))

		# First check if all members are present
		present_keys = set(values.keys())
		missing_keys = self._required_keys - present_keys
		if len(missing_keys) > 0:
			raise ProgrammerErrorException("Missing keys in %s: %s" % (self.__class__.__name__, ", ".join(sorted(missing_keys))))

		result_data = bytearray()
		for member in self.members:
			if not member.implicit_value:
				value = values[member.name]
				result_data += member.pack(value)
			else:
				# Data is implicit
				result_data += member.pack()
		return bytes(result_data)

	def _unpack(self, databuffer):
		result = { }
		with databuffer.rewind_on_exception():
			for member in self.members:
				value = member.unpack(databuffer)
				if not member.implicit_value:
					result[member.name] = value
			return result

	def __repr__(self):
		if self.name is None:
			return "%s" % (self.typename)
		else:
			return "%s(%s)" % (self.typename, self.name)

	def __str__(self):
		return "%s<%s>" % (self.name, ", ".join(("%s %s" % (member.name, member.typename) for member in self.members)))

class VariableType(BaseStructureMember):
	def __init__(self, possible_inner_classes, name = None):
		BaseStructureMember.__init__(self, name = name)
		self._possible_inner_classes = possible_inner_classes

	@property
	def typename(self):
		return "Variable"

	def pack(self, type_value_tuple):
		if not isinstance(type_value_tuple, tuple):
			raise InvalidInputTypeException("%s requires a tuple to be supplied for packing, got %s: %s" % (self.typename, type(type_value_tuple).__name__, str(type_value_tuple)))
		if len(type_value_tuple) != 2:
			raise InvalidInputTypeException("%s requires a 2-tuple to be supplied for packing, got %d-tuple: %s" % (self.typename, len(type_value_tuple), str(type_value_tuple)))
		(inner_class, values) = type_value_tuple
		return inner_class.pack(values)

	def _unpack(self, databuffer):
		for possible_inner_class in self._possible_inner_classes:
			try:
				unpacking = possible_inner_class.unpack(databuffer)
			except (DeserializationException, DataBufferException):
				continue
			return (possible_inner_class, unpacking)
		raise NoPossibleDeserializationException("Attempts to deserialize all possible %d inner classes were unsuccessful." % (len(self._possible_inner_classes)))

def instantiate_member(name, typename, **kwargs):
	return StructureMemberFactory.instantiate(name, typename, **kwargs)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2021 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
import datetime
import tempfile
import enum
import pyasn1.codec.der.encoder
from pyasn1_modules import rfc2459
from x509sak.SubprocessExecutor import SubprocessExecutor
from x509sak.PEMDERObject import PEMDERObject
from x509sak.DistinguishedName import DistinguishedName
from x509sak.Tools import ASN1Tools
from x509sak.AlgorithmDB import SignatureAlgorithms
from x509sak.OID import OID, OIDDB
from x509sak.PublicKey import PublicKey
from x509sak.X509Extensions import X509ExtensionRegistry, X509Extensions
from x509sak.Exceptions import UnknownAlgorithmException
from x509sak.OpenSSLTools import OpenSSLTools

_log = logging.getLogger("x509sak.X509Certificate")

class X509CertificateClass(enum.IntEnum):
	CARoot = 1
	CAIntermediate = 2
	ClientServerAuth = 3
	ServerAuth = 4
	ClientAuth = 5
	Other = 6

class X509Certificate(PEMDERObject):
	_PEM_MARKER = "CERTIFICATE"
	_ASN1_MODEL = rfc2459.Certificate

	def _post_decode_hook(self):
		self._extensions = None

	@property
	def signed_payload(self):
		return pyasn1.codec.der.encoder.encode(self.asn1["tbsCertificate"])

	@property
	def signature_algorithm(self):
		algorithm_oid = OID.from_asn1(self.asn1["signatureAlgorithm"]["algorithm"])
		sig_alg = SignatureAlgorithms.lookup("oid", algorithm_oid)
		if sig_alg is None:
			raise UnknownAlgorithmException("Unknown signature OID %s." % (algorithm_oid))
		return sig_alg

	@property
	def version(self):
		return int(self.asn1["tbsCertificate"]["version"]) + 1

	@property
	def serial(self):
		return int(self.asn1["tbsCertificate"]["serialNumber"])

	@property
	def pubkey(self):
		pubkey_asn1 = self.asn1["tbsCertificate"]["subjectPublicKeyInfo"]
		return PublicKey.from_asn1(pubkey_asn1)

	@property
	def subject(self):
		return DistinguishedName.from_asn1(self.asn1["tbsCertificate"]["subject"])

	@property
	def issuer(self):
		return DistinguishedName.from_asn1(self.asn1["tbsCertificate"]["issuer"])

	@property
	def valid_not_before(self):
		return ASN1Tools.parse_datetime(str(self.asn1["tbsCertificate"]["validity"]["notBefore"].getComponent()))

	@property
	def valid_not_after(self):
		return ASN1Tools.parse_datetime(str(self.asn1["tbsCertificate"]["validity"]["notAfter"].getComponent()))

	@property
	def subject_unique_id(self):
		return self.asn1["tbsCertificate"]["subjectUniqueID"].asInteger() if self.asn1["tbsCertificate"]["subjectUniqueID"].hasValue() else None

	@property
	def issuer_unique_id(self):
		return self.asn1["tbsCertificate"]["issuerUniqueID"].asInteger() if self.asn1["tbsCertificate"]["issuerUniqueID"].hasValue() else None

	@property
	def signature_alg_oid(self):
		signature_alg_oid = OID.from_str(str(self.asn1["signatureAlgorithm"]["algorithm"]))
		return signature_alg_oid

	@property
	def signature_alg_params(self):
		if self.asn1["signatureAlgorithm"]["parameters"].hasValue():
			return bytes(self.asn1["signatureAlgorithm"]["parameters"])
		else:
			return None

	@property
	def signature(self):
		signature = ASN1Tools.bitstring2bytes(self.asn1["signatureValue"])
		return signature

	@property
	def extensions(self):
		if self._extensions is None:
			self._extensions = self._get_extensions()
		return self._extensions

	def is_time_valid(self, now = None):
		if now is None:
			now = datetime.datetime.utcnow()
		return self.valid_not_before <= now <= self.valid_not_after

	def seconds_until_expires(self, now = None):
		if now is None:
			now = datetime.datetime.utcnow()
		return (self.valid_not_after - now).total_seconds()

	def _get_extensions(self):
		result = [ ]
		if self.asn1["tbsCertificate"]["extensions"] is not None:
			for extension in self.asn1["tbsCertificate"]["extensions"]:
				oid = OID.from_asn1(extension["extnID"])
				critical = bool(extension["critical"])
				value = bytes(extension["extnValue"])
				result.append(X509ExtensionRegistry.create(oid, critical, value))
		return X509Extensions(result)

	@property
	def is_selfsigned(self):
		return self.signed_by(self)

	def signed_by(self, potential_issuer, verbose_failure = False):
		return OpenSSLTools.validate_signature(subject_certificate = self, issuer_certificate = potential_issuer, verbose_failure = verbose_failure)

	def dump_pem(self, f = None):
		print("# Subject : %s" % (self.subject.pretty_str), file = f)
		print("# Issuer  : %s" % (self.issuer.pretty_str), file = f)
		print("# Validity: %s UTC - %s UTC" % (self.valid_not_before.strftime("%Y-%m-%d %H:%M:%S"), self.valid_not_after.strftime("%Y-%m-%d %H:%M:%S")), file = f)
		print("# Hash    : %s" % (self.hashval.hex()), file = f)
		print(self.to_pem_data(), file = f)
		print(file = f)

	@property
	def is_ca_certificate(self):
		if len(self.extensions) == 0:
			return True
		else:
			# RFC5280, 4.2: "A certificate MUST NOT include more than one
			# instance of a particular extension." -- we assume the certificate
			# is sane and elect to get the first extension.
			basic_constraints = self.extensions.get_first(OIDDB.X509Extensions.inverse("BasicConstraints"))
			if basic_constraints is None:
				# RFC5280, 4.2.1.9: "If the basic constraints extension is not present in a
				# version 3 certificate, or the extension is present but the cA boolean
				# is not asserted, then the certified public key MUST NOT be used to
				# verify certificate signatures.
				if self.version < 3:
					return True
				else:
					return False
			else:
				return basic_constraints.is_ca

	def classify(self):
		if self.is_ca_certificate:
			if self.is_selfsigned:
				return X509CertificateClass.CARoot
			else:
				return X509CertificateClass.CAIntermediate
		else:
			eku = self.extensions.get_first(OIDDB.X509Extensions.inverse("ExtendedKeyUsage"))
			if eku is not None:
				(client, server) = (eku.client_auth, eku.server_auth)
				if client and server:
					return X509CertificateClass.ClientServerAuth
				elif client:
					return X509CertificateClass.ClientAuth
				elif server:
					return X509CertificateClass.ServerAuth
			return X509CertificateClass.Other

	@classmethod
	def from_tls_server(cls, hostname, port = 443):
		pem_data = OpenSSLTools.get_tls_server_cert_pem(hostname = hostname, port = port)
		certificates = cls.from_pem_data(pem_data)
		return certificates[0]

	def dump(self):
		print("Subject:")
		self.subject.dump(indent = "    ")
		print()
		print("Issuer:")
		self.issuer.dump(indent = "    ")
		print()
		self.extensions.dump()

	def __str__(self):
		return "X509Certificate<subject = %s, issuer = %s>" % (self.subject.rfc2253_str, self.issuer.rfc2253_str)
